/**
*	This sample redirects DNS requests to opendns.com server.
**/

#include "stdafx.h"
#include <crtdbg.h>
#include <map>
#include "nfapi.h"
#include "nfutil.h"

using namespace nfapi;

// Change this string after renaming and registering the driver under different name
#define NFDRIVER_NAME "netfilter2"

// Forward declarations
void printAddrInfo(bool created, ENDPOINT_ID id, PNF_UDP_CONN_INFO pConnInfo);

//
//	API events handler
//
class EventHandler : public NF_EventHandler
{
	struct ORIGINAL_CONN_INFO
	{
		unsigned char	remoteAddress[NF_MAX_ADDRESS_LENGTH];
	};

	typedef std::map<ENDPOINT_ID, ORIGINAL_CONN_INFO> tConnInfoMap;
	tConnInfoMap m_connInfoMap;


	virtual void threadStart()
	{
	}

	virtual void threadEnd()
	{
	}
	
	//
	// TCP events
	//
	virtual void tcpConnectRequest(ENDPOINT_ID id, PNF_TCP_CONN_INFO pConnInfo)
	{
	}

	virtual void tcpConnected(ENDPOINT_ID id, PNF_TCP_CONN_INFO pConnInfo)
	{
	}

	virtual void tcpClosed(ENDPOINT_ID id, PNF_TCP_CONN_INFO pConnInfo)
	{
	}

	virtual void tcpReceive(ENDPOINT_ID id, const char * buf, int len)
	{	
	}

	virtual void tcpSend(ENDPOINT_ID id, const char * buf, int len)
	{
	}

	virtual void tcpCanReceive(ENDPOINT_ID id)
	{
	}

	virtual void tcpCanSend(ENDPOINT_ID id)
	{
	}
	
	//
	// UDP events
	//

	virtual void udpCreated(ENDPOINT_ID id, PNF_UDP_CONN_INFO pConnInfo)
	{
		printAddrInfo(true, id, pConnInfo);
		fflush(stdout);
	}

	virtual void udpConnectRequest(ENDPOINT_ID id, PNF_UDP_CONN_REQUEST pConnReq)
	{
		char remoteAddr[MAX_PATH];
		DWORD dwLen;
		
		dwLen = sizeof(remoteAddr);
		WSAAddressToString((sockaddr*)pConnReq->remoteAddress, 
				(((sockaddr*)pConnReq->remoteAddress)->sa_family == AF_INET6)? sizeof(sockaddr_in6) : sizeof(sockaddr_in), 
				NULL, 
				remoteAddr, 
				&dwLen);

		// Redirect only IPv4 DNS requests
		if (((sockaddr*)pConnReq->remoteAddress)->sa_family == AF_INET)
		{
			sockaddr_in * pAddr = (sockaddr_in*)pConnReq->remoteAddress;
	
			if (pAddr->sin_port == htons(53)) // DNS port
			{
				ORIGINAL_CONN_INFO oci;

				memcpy(oci.remoteAddress, pConnReq->remoteAddress, sizeof(oci.remoteAddress));

				// Save original remote address
				m_connInfoMap[id] = oci;

				// Redirect DNS request to opendns.com
				pAddr->sin_addr.S_un.S_addr = inet_addr("208.67.222.220");
				
				printf("udpConnectRequest id=%I64u remoteAddress=%s - DNS request, redirected to opendns.com\n", id, remoteAddr);
				fflush(stdout);
				return;
			}
		}		

		printf("udpConnectRequest id=%I64u remoteAddress=%s\n", id, remoteAddr);
		fflush(stdout);
	}

	virtual void udpClosed(ENDPOINT_ID id, PNF_UDP_CONN_INFO pConnInfo)
	{
		printAddrInfo(false, id, pConnInfo);
		fflush(stdout);

		m_connInfoMap.erase(id);
	}

	virtual void udpReceive(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, PNF_UDP_OPTIONS options)
	{	
		char remoteAddr[MAX_PATH];
		DWORD dwLen;
		
		dwLen = sizeof(remoteAddr);
		WSAAddressToString((sockaddr*)remoteAddress, 
				(((sockaddr*)remoteAddress)->sa_family == AF_INET6)? sizeof(sockaddr_in6) : sizeof(sockaddr_in), 
				NULL, 
				remoteAddr, 
				&dwLen);

		printf("udpReceive id=%I64u len=%d remoteAddress=%s\n", id, len, remoteAddr);
		fflush(stdout);

		tConnInfoMap::iterator it = m_connInfoMap.find(id);
		if (it != m_connInfoMap.end())
		{
			// Restore original remote address
			nf_udpPostReceive(id, it->second.remoteAddress, buf, len, options);
			return;
		}

		// Send the packet to application
		nf_udpPostReceive(id, remoteAddress, buf, len, options);
	}

	virtual void udpSend(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, PNF_UDP_OPTIONS options)
	{
		char remoteAddr[MAX_PATH];
		DWORD dwLen;
		
		dwLen = sizeof(remoteAddr);
		WSAAddressToString((sockaddr*)remoteAddress, 
				(((sockaddr*)remoteAddress)->sa_family == AF_INET6)? sizeof(sockaddr_in6) : sizeof(sockaddr_in), 
				NULL, 
				remoteAddr, 
				&dwLen);

		// Redirect only IPv4 DNS requests
		if (((sockaddr*)remoteAddress)->sa_family == AF_INET)
		{
			sockaddr_in addr = *(sockaddr_in*)remoteAddress;
	
			if (addr.sin_port == htons(53)) // DNS port
			{
				ORIGINAL_CONN_INFO oci;

				memcpy(oci.remoteAddress, remoteAddress, sizeof(oci.remoteAddress));

				// Save original remote address
				m_connInfoMap[id] = oci;

				// Redirect DNS request to opendns.com
				addr.sin_addr.S_un.S_addr = inet_addr("208.67.222.222");
				
				// Post the redirected packet
				nf_udpPostSend(id, (unsigned char *)&addr, buf, len, options);
				
				printf("udpSend id=%I64u len=%d remoteAddress=%s - DNS request, redirected to opendns.com\n", id, len, remoteAddr);
				fflush(stdout);
				return;
			}
		}		

		printf("udpSend id=%I64u len=%d remoteAddress=%s\n", id, len, remoteAddr);
		fflush(stdout);

		// Send the packet to server
		nf_udpPostSend(id, remoteAddress, buf, len, options);
	}

	virtual void udpCanReceive(ENDPOINT_ID id)
	{
	}

	virtual void udpCanSend(ENDPOINT_ID id)
	{
	}
};

int main(int argc, char* argv[])
{
	EventHandler eh;
	NF_RULE rule;
	WSADATA wsaData;

	// This call is required for WSAAddressToString
    ::WSAStartup(MAKEWORD(2, 2), &wsaData);

#ifdef _DEBUG
	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
	_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_DEBUG);
#endif

	printf("Press any key to stop...\n\n");

	// Initialize the library and start filtering thread
	if (nf_init(NFDRIVER_NAME, &eh) != NF_STATUS_SUCCESS)
	{
		printf("Failed to connect to driver");
		return -1;
	}

	// Filter outgoing IPv4 UDP packets, directed to port 53 (DNS)
	memset(&rule, 0, sizeof(rule));
	rule.ip_family = AF_INET;
	rule.protocol = IPPROTO_UDP;
	rule.direction = NF_D_OUT;
	rule.remotePort = htons(53);
	rule.filteringFlag = NF_FILTER | NF_INDICATE_CONNECT_REQUESTS;
	nf_addRule(&rule, TRUE);

	memset(&rule, 0, sizeof(rule));
	rule.ip_family = AF_INET;
	rule.protocol = IPPROTO_UDP;
	rule.direction = NF_D_IN;
	rule.filteringFlag = NF_FILTER;
	nf_addRule(&rule, TRUE);

	// Wait for any key
	getchar();

	// Free the library
	nf_free();

	::WSACleanup();

	return 0;
}


/**
* Print the address information
**/
void printAddrInfo(bool created, ENDPOINT_ID id, PNF_UDP_CONN_INFO pConnInfo)
{
	char localAddr[MAX_PATH] = "";
	sockaddr * pAddr;
	DWORD dwLen;
	char processName[MAX_PATH] = "";
	
	pAddr = (sockaddr*)pConnInfo->localAddress;
	dwLen = sizeof(localAddr);

	WSAAddressToString((LPSOCKADDR)pAddr, 
				(pAddr->sa_family == AF_INET6)? sizeof(sockaddr_in6) : sizeof(sockaddr_in), 
				NULL, 
				localAddr, 
				&dwLen);
		
	if (created)
	{
		if (!nf_getProcessName(pConnInfo->processId, processName, sizeof(processName)/sizeof(processName[0])))
		{
			processName[0] = '\0';
		}

		printf("udpCreated id=%I64u pid=%d local=%s\n\tProcess: %s\n",
			id,
			pConnInfo->processId, 
			localAddr, 
			processName);
	} else
	{
		printf("udpClosed id=%I64u pid=%d local=%s\n",
			id,
			pConnInfo->processId, 
			localAddr);
	}

}
